'use strict';

exports.__esModule = true;

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

exports.combineReducers = combineReducers;
exports.mapValues = mapValues;
exports.createReducer = createReducer;

function defaultEqualityCheck(a, b) {
  return a === b;
}

function defaultMemoize(func) {
  var equalityCheck = arguments.length <= 1 || arguments[1] === undefined ? defaultEqualityCheck : arguments[1];

  var lastArgs = null;
  var lastResult = null;
  return function () {
    for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    if (lastArgs !== null && args.every(function (value, index) {
      return equalityCheck(value, lastArgs[index]);
    })) {
      return lastResult;
    }
    lastArgs = args;
    lastResult = func.apply(undefined, args);
    return lastResult;
  };
}

function pick(obj, fn) {
  return Object.keys(obj).reduce(function (result, key) {
    if (fn(obj[key])) {
      result[key] = obj[key];
    }
    return result;
  }, {});
}

function combineReducers(reducers) {
  var finalReducers = pick(reducers, function (val) {
    return typeof val === 'function';
  });
  return function (state, action) {
    if (state === undefined) state = {};
    return mapValues(finalReducers, function (reducer, key) {
      return reducer(state[key], action);
    });
  };
}

function mapValues(obj, fn) {
  return Object.keys(obj).reduce(function (result, key) {
    result[key] = fn(obj[key], key);
    return result;
  }, {});
}

function parseArgs(args) {
  if (args.length == 1) {
    return {
      reducerDependencies: {},
      reducer: args[0],
      reducerMemoizer: defaultMemoize // TODO allow customization
    };
  } else if (args.length == 2) {
      return {
        reducerDependencies: args[0],
        reducer: args[1],
        reducerMemoizer: defaultMemoize
      };
    } else {
      throw new Error('Bad usage'); // TODO better errors
    }
}

function wrapState(state, dependencies) {
  if (typeof state === 'object') {
    if (typeof state.__dependencies !== 'undefined') {
      throw new Error('__dependencies is a reserved state attribute name');
    }
    return _extends({}, state, {
      __dependencies: dependencies
    });
  } else {
    return {
      value: state,
      __dependencies: dependencies
    };
  }
}

function unwrapState(state) {
  if (Object.keys(state).length === 2 && state.__dependencies && typeof state.value !== 'undefined') {
    return state.value;
  } else {
    return state;
  }
}

function createReducer() {
  var _parseArgs = parseArgs(arguments);

  var reducerDependencies = _parseArgs.reducerDependencies;
  var reducer = _parseArgs.reducer;
  var reducerMemoizer = _parseArgs.reducerMemoizer;

  var hasDependencies = Object.keys(reducerDependencies).length > 0;

  if (!hasDependencies) {
    return reducerMemoizer(reducer);
  }

  var reducerDependenciesReducer = reducerMemoizer(combineReducers(reducerDependencies));

  var liftedReducer = function liftedReducer(state, action) {
    if (state === undefined) state = { value: undefined, __dependencies: undefined };
    var value = state.value;
    var __dependencies = state.__dependencies;

    var nextDependenciesState = reducerDependenciesReducer(__dependencies, action);

    var nextDependenciesStatesUnwrapped = mapValues(nextDependenciesState, function (depState) {
      return unwrapState(depState);
    });

    var nextReducerState = reducer(value, action, nextDependenciesStatesUnwrapped);

    if (typeof nextReducerState === 'object' && typeof nextReducerState.__dependencies !== 'undefined') {
      throw new Error('A reducer that depends on other reducers should not return an object with the __dependencies' + ' attribute as it is a reserved attribute name used by the library');
    }

    return wrapState(nextReducerState, nextDependenciesState);
  };

  return reducerMemoizer(liftedReducer);
}